#include "elemlist"

template<typename collitemtype, class Op>
class Group3test
{
    collitemtype base;
    const Lista<Op> &list;
public:
    Group3test(collitemtype b, const Lista<Op> &l) : base(b), list(l) { }
    static unsigned groupcount() { return Op::groupcount(); }
    bool needsparm(collitemtype c) const { return list[c-base].needsparm(); }
    bool compatible(collitemtype a, collitemtype b) const { return list[a-base].compatible(list[b-base]); }
};

#include <list>
#include <vector>
#include <pair.h>
using std::list;
using std::pair;
using std::vector;

template<typename collitemtype, class Op>
class JobList
{
    typedef typename std::set<collitemtype>::const_iterator it;
    
    collitemtype agbase, chbase, mdbase;
    const Lista<Agent> &aglist;
    const Lista<Channel> &chlist;
    const Lista<Op> &mdlist;
public:
    typedef list<pair<string, vector<collitemtype> > > joblisttype;
    joblisttype joblist;

    JobList(collitemtype a, collitemtype c, collitemtype m,
              const Lista<Agent> &ag, const Lista<Channel> &ch,
              const Lista<Op> &md) : agbase(a), chbase(c), mdbase(m),
                aglist(ag), chlist(ch), mdlist(md) { }
    void operator()
    (
      const it ag1, const it ag2,
      const it ch1, const it ch2,
      const it md1, const it md2
    )
    {
        string s;
        for(it i = ch1; i != ch2; ++i) { if(s.size())s += ','; s += chlist[*i-chbase].getname(); }
        if(s.size())s += ' ';
        
        list<Op> ops;
        for(it i = md1; i != md2; ++i)
            ops.push_back(mdlist[*i-mdbase]);
        
        s += strgroup(ops);
        
        vector<collitemtype> agents(ag1, ag2);
        for(unsigned a=0; a<agents.size(); ++a)agents[a] -= agbase;
        
        joblist.push_back(make_pair(s, agents));
    }
    
    void dump() const
    {
        typename joblisttype::const_iterator i;
        for(i=joblist.begin(); i!=joblist.end(); ++i)
        {
            cout << "# Agents:";
            typename vector<collitemtype>::const_iterator j;
            for(j=i->second.begin(); j!=i->second.end(); ++j)
            {
                const Agent &a = aglist[*j];
                cout << ' ';
                a.dump();
            }
            cout << "\n# Job: \"" << i->first << "\"\n# \n";
        }
    }
    
    void distribute()
    {
#ifdef ELEMLIST_DEBUG
        dump();
#endif
        typedef typename joblisttype::const_iterator tit;
        
        unsigned combcount=1;        
        for(tit i=joblist.begin(); i!=joblist.end(); ++i)
            combcount *= i->second.size();
        
        vector<unsigned> loads(aglist.Size());
        for(unsigned b=0; b<loads.size(); ++b)
            loads[b] = aglist[b].getload();
        
        unsigned bestload = 0, leastload = 0, bestcomb=0, relying=0;
        for(unsigned a=0; a<combcount; ++a)
        {
            vector<unsigned> tmploads = loads;
            unsigned c=a;
            for(tit i=joblist.begin(); i!=joblist.end(); ++i)
            {
                unsigned mod = i->second.size();
                unsigned agnum = c%mod; c/=mod;
                ++tmploads[i->second[agnum]];
            }
            c=a;
            unsigned loadsum=0, relysum = 0;
            for(tit i=joblist.begin(); i!=joblist.end(); ++i)
            {
                unsigned mod = i->second.size();
                unsigned agnum = c%mod; c/=mod;
                loadsum += tmploads[i->second[agnum]];
                relysum += aglist[i->second[agnum]].getrely();
            }

            unsigned maxload = 0;
            for(unsigned b=0; b<tmploads.size(); ++b)
                if(tmploads[b] > maxload) maxload = tmploads[b];
            
#ifdef ELEMLIST_DEBUG
            cout << "# Choice " << a << ": Load would be " << maxload << " (" << loadsum << ")";
#endif
            if(maxload < bestload || !a)
            {
                bestload = maxload;
            Good:
                relying = relysum;
                leastload= loadsum;
                bestcomb = a; 
#ifdef ELEMLIST_DEBUG
                cout << " - using";
#endif
            }
            else if(maxload == bestload && loadsum < leastload || relysum > relying)
                goto Good;
           
#ifdef ELEMLIST_DEBUG
            cout << endl;
#endif
        }
        
        for(tit i=joblist.begin(); i!=joblist.end(); ++i)
        {
            unsigned mod = i->second.size();
            unsigned agnum = bestcomb%mod; bestcomb/=mod;
            
            cout << Op::cmdchar() << ' ' << aglist[i->second[agnum]].getname() << ' ' << i->first << endl;
        }
    }
};

template<class Op>
void ActMap<Op>::Distribute()
{
#ifdef ELEMLIST_DEBUG
    typedef char collitemtype;
    collitemtype aBase = 'A';
    collitemtype cBase = 'A' + As();
    collitemtype mBase = 'A' + As() + Cs();
    
    cout << "# Agents: ";
    for(unsigned a=0; a<As(); ++a) { cout << (collitemtype)(a+aBase) << '('; Agents[a].dump(); cout << ") "; }
    cout << "\n# Channels: ";
    for(unsigned c=0; c<Cs(); ++c) { cout << (collitemtype)(c+cBase) << '('; Channels[c].dump(); cout << ") "; }
    cout << "\n# Ops: ";
    for(unsigned m=0; m<Ms(); ++m) { cout << (collitemtype)(m+mBase) << '('; Ops[m].dump(); cout << ") "; }
    cout << endl;
#else
    typedef unsigned collitemtype;
    collitemtype aBase=0, cBase=0, mBase=0;
#endif

    typedef elemlist<3, collitemtype> colltype;
    colltype coll;
    
    for(unsigned m=0; m<Ms(); ++m)
        for(unsigned a=0; a<As(); ++a)
            for(unsigned c=0; c<Cs(); ++c)
                if(map[MdInd(a,c,m)])
                {
                    colltype::Elem tmp;
                    tmp.insert(0, (collitemtype)(a+aBase));
                    tmp.insert(1, (collitemtype)(c+cBase));
                    tmp.insert(2, (collitemtype)(m+mBase));
                    coll.insert(tmp);
                }
#ifdef ELEMLIST_DEBUG
    cout << "# Unoptimized: ";
    coll.dump();
    cout << endl;
#endif
    coll.optimize();
#ifdef ELEMLIST_DEBUG
    cout << "# Optimized: ";
    coll.dump();
    cout << endl;
#endif
    coll.expand3(Group3test<collitemtype,Op> (mBase, Ops));
#ifdef ELEMLIST_DEBUG
    cout << "# Expanded: ";
    coll.dump();
    cout << endl;
#endif
    coll.crop();
#ifdef ELEMLIST_DEBUG
    cout << "# Cropped: ";
    coll.dump();
    cout << endl;
#endif
    
    JobList<collitemtype, Op> jobs(
      aBase,cBase,mBase,
      Agents,Channels,Ops);
    coll.format(jobs);
    
#ifdef ELEMLIST_DEBUG
    cout << "# ";
    coll.dump();
    cout << endl;
#endif

    jobs.distribute();
}
template<class Op>
void ActMap<Op>::SetChanOp(const Channel &ch, const Op &md)
{
    unsigned mi = OpIndex(md);
    unsigned ci = ChannelIndex(ch);
    bool got = DoMarkOp(ci, mi);
    if(!got)
    {
        cout << "W " << ch.getname() << ' ';
        ch.dump();
        cout << ' ';
        md.dump(); 
        cout << endl;
    }
}
